import json
import hashlib
from speckle.base.resource import ResourceBase
from pydantic import BaseModel, validator
from typing import List, Optional
from speckle.base.resource import ResourceBase, ResourceBaseSchema

NAME = 'objects'
METHODS = ['list', 'get', 'update', 'create'
           'delete', 'comment_get', 'comment_create',
           'get_bulk', 'set_properties']


class SpeckleObject(ResourceBaseSchema):
    type: Optional[str]
    name: Optional[str]
    geometryHash: Optional[str]  # Is immediately replaced anyways
    hash: Optional[str]  # Is immediately replaced anyways
    applicationId: Optional[str]
    properties: dict = {}
    partOf: Optional[List[str]]
    parent: Optional[List[str]]
    children: Optional[List[str]]
    ancestors: Optional[List[str]]


    def dict(self):
        json_string = json.dumps(super(SpeckleObject, self).dict()['properties'])

        self.geometryHash = hashlib.md5(
            json_string.encode('utf-8')).hexdigest()

        self.hash = hashlib.md5('{}.{}'.format(self.type, json_string).encode('utf-8')).hexdigest()

        return super(SpeckleObject, self).dict()


class Resource(ResourceBase):
    """API Access class for Speckle Objects

    """
    
    def __init__(self, session, basepath, me):
        super().__init__(session, basepath, me, NAME, METHODS)

        self.method_dict.update({
            'get_bulk': {
                'method': 'POST'
            },
            'set_properties': {
                'method': 'PUT'
            }
        })
           
        self.schema = SpeckleObject

    def list(self):
        """List all Speckle objects
        
        Returns:
            list -- A list of Speckle object data class instances
        """
        return self.make_request('list', '/')

    def create(self, data):
        """Create a Speckle object from a data dictionary
        
        Arguments:
            data {dict} -- A dictionary describing a Speckle object
        
        Returns:
            SpeckleObject -- The instance created on the Speckle Server
        """
        return self.make_request('create', '/', data)

    def get(self, id):
        """Get a specific Speckle object from the SpeckleServer
        
        Arguments:
            id {str} -- The ID of the Speckle object to retrieve
        
        Returns:
            SpeckleObject -- The Speckle object
        """
        return self.make_request('get', '/' + id)

    def update(self, id, data):
        """Update a specific Speckle object
        
        Arguments:
            id {str} -- The ID of the Speckle object to update
            data {dict} -- A dict of values to update
        
        Returns:
            dict -- a confirmation payload with the updated keys
        """
        return self.make_request('update', '/' + id, data)

    def delete(self, id):
        """Delete a specific Speckle object
        
        Arguments:
            id {str} -- The ID of the Speckle object to delete
        
        Returns:
            dict -- A confirmation payload
        """
        return self.make_request('delete', '/' + id)

    def comment_get(self, id):
        """Retrieve comments attached to a Speckle object
        
        Arguments:
            id {str} -- The ID of the Speckle object to retrieve comments from
        
        Returns:
            list -- A list of comments
        """
        return self.make_request('comment_get', '/' + id, comment=True)

    def comment_create(self, id, data):
        """Add a comment to a Speckle object
        
        Arguments:
            id {str} -- The ID of the Speckle object to comment on
            data {dict} -- A comment dictionary object
        
        Returns:
            CommentSchema -- The comment created by the server
        """
        return self.make_request('comment_create', '/' + id, data, comment=True)

    def get_bulk(self, object_ids, query):
        """Retrieve and optionally update a list of Speckle objects
        
        Arguments:
            object_ids {list} -- A list of object IDs
            query {dict} -- A dictionary to specifiy which fields to retrieve, filters, limits, etc
        
        Returns:
            list -- A list of SpeckleObjects
        """
        query_string = '?'

        for key, value in query.items:
            query_string += key + '='
            if isinstance(value, list):
                query_string += ','.join(value)
            elif isinstance(value, str):
                query_string += value + '&'
            else:
                raise 'query dict values must be list or string but key {} is of type {}'.format(key, type(value))

        query_string = query_string[:-1] # Remove last '&' or '?' to be clean

        return self.make_request('get_bulk', '/getbulk' + query_string, object_ids)

    def set_properties(self, id, data):
        return self.make_request('set_properties', '/' + id + '/properties', data)



import uuid
import pytest


@pytest.fixture(scope='module')
def resource():
    return {
        'name': 'example stream',
        'description': 'a stream for testing purposes',
        'tags': [
            'test',
            'example'
        ],
        'commitMessage': 'created stream',
        'objects': [],
        'layers': [],
    }


@pytest.fixture(scope='module')
def updated_resource():
    return {
        'name': 'updated stream',
        'description': 'a stream for testing purposes',
        'tags': [
            'test',
            'example'
        ],
        'commitMessage': 'created stream',
        'objects': [],
        'layers': [],
    }


@pytest.fixture(scope='module')
def created_item(client, resource):
    return client.streams.create(resource)


@pytest.fixture
def comment():
    return {
        'text': 'Wow this is an amazing resource!',
    }

@pytest.fixture()
def api_client(client, created_item):
    api_client_dict = {
        'role': 'Hybrid',
        'documentName': 'Test',
        'documentType': 'DataServer',
        'documentLocation': 'http://some.server.com',
        'documentGuid': 'some-long-uuid',
        'online': True,
        'streamId': created_item.streamId
    }
    api_client = client.api_clients.create(api_client_dict)
    yield api_client
    client.api_clients.delete(api_client.id)

def test_create(client, resource):
    returned_resource = client.streams.create(resource)

    assert resource['name'] == returned_resource.name
    assert resource['description'] == returned_resource.description
    assert resource['tags'] == returned_resource.tags
    assert resource['objects'] == returned_resource.objects
    assert resource['layers'] == returned_resource.layers

    assert returned_resource.streamId is not None


@pytest.mark.dependency(depends=['test_create'])
def test_list(client, created_item):
    streams = client.streams.list()

    assert len(streams) == 2


@pytest.mark.dependency(depends=['test_create'])
def test_get(client, created_item):
    stream = client.streams.get(created_item.streamId)

    assert stream.id == created_item.id


@pytest.mark.dependency(depends=['test_create'])
def test_clone(client, created_item):
    clone, parent = client.streams.clone(created_item.streamId, 'cloned_stream')
    
    assert clone.name == 'cloned_stream'
    assert parent.name == created_item.name

    assert clone.parent == parent.streamId
    assert parent.children == [clone.streamId]


@pytest.mark.dependency(depends=['test_create'])
def test_list_clients_empty(client, created_item):
    api_clients = client.streams.list_clients(created_item.streamId)

    assert api_clients == []


@pytest.mark.dependency(depends=['test_create'])
def test_list_clients(client, created_item, api_client):
    api_clients = client.streams.list_clients(created_item.streamId)

    assert api_clients[0].id == api_client.id

def test_list_objects(client):
    # TODO(): add list objects test
    pass


def test_diff(client):
    # TODO(): add diff streams
    pass


@pytest.mark.dependency(depends=['test_create'])
def test_comments(client, created_item, comment):
    comments = client.streams.comment_get(created_item.streamId)
    assert comments == []
    client.streams.comment_create(created_item.streamId, comment)
    comments = client.streams.comment_get(created_item.streamId)
    assert len(comments) == 1
    assert comments[0].text == comment['text']

@pytest.mark.dependency(depends=['test_create'])
def test_delete(client, resource):
    data = client.streams.list()
    assert data != []
    for stream in data:
        client.streams.delete(id=stream.streamId)



import uuid
from speckle.base.resource import ResourceBase
from pydantic import BaseModel, UUID4, validator
from typing import List, Optional, Union
from speckle.base.resource import ResourceBase, ResourceBaseSchema
from speckle.resources.accounts import User
from enum import Enum

NAME = 'clients'
METHODS = ['list', 'create', 'get', 'update',
           'delete']


# class RoleEnum(str, Enum):
#     receiver: 'Receiver'
#     sender: 'Sender'
#     hybrid: 'Hybrid'


class ApiClient(ResourceBaseSchema):
    # role: Optional[RoleEnum] = RoleEnum.receiver
    role: Optional[str]
    documentName: Optional[str]
    documentType: Optional[str]
    documentLocation: Optional[str]
    documentGuid: Optional[str]
    streamId: Optional[str]
    online: Optional[bool]
    owner: Optional[Union[User, str]]

    @validator('documentGuid', pre=True, always=True)
    def set_guid(cls, v):
        return v or str(uuid.uuid4())

class Resource(ResourceBase):
    """API Access class for API Clients

    The `api_client` resource mostly returns an API Client data class instance.

    Example:
        Here is an example of what an API Client object looks like in dict form and when converted to a data class: 
        
        .. code-block:: python

            >>> from speckle.resources.api_clients import ApiClient
            >>> api_client_dict =  {
                    'role': 'Hybrid',
                    'documentName': 'Test',
                    'documentType': 'DataServer',
                    'documentLocation': 'http://some.server.com',
                    'online': True,
                }
            >>>  api_client = ApiClient.parse_obj(api_client_dict)
            >>> api_client.role
            'Hybrid'
            >>> api_client.documentGuid # Automatically created when document is instantiated if not specified
            'e991c923-cbb7-45f4-9488-e0f61ba006c0'



    """
    def __init__(self, session, basepath, me):
        super().__init__(session, basepath, me, NAME, METHODS)

        self.schema = ApiClient

    def list(self):
        """List all API clients
        
        Returns:
            list -- A list of API client data class instances

    Example:
        Calling the `api_clients.list()` method returns a list of ApiClient class objects that the user has read access to.
        .. code-block:: python

            >>> client.api_clients.list()
            [<ApiClient>, <ApiClient>, <ApiClient>]
        """
        return self.make_request('list', '/')

    def create(self, data):
        """Create an API clients from a data dictionary
        
        Arguments:
            data {dict} -- A dictionary describing an API client
        
        Returns:
            ApiClient -- The instance created on the Speckle Server

    Example:        
        .. code-block:: python

            >>> api_client_dict =  {
                    'role': 'Hybrid',
                    'documentName': 'Test',
                    'documentType': 'DataServer',
                    'documentLocation': 'http://some.server.com',
                    'online': True,
                }
            >>> new_api_client = client.api_clients.create(data=api_client_dict)
            >>> new_api_client.id
            '54759eb3c090d83494e2d804'
            >>> new_api_client.documentName
            'Test'
        """
        return self.make_request('create', '/', data)

    def get(self, id):
        """Get a specific API client from the SpeckleServer
        
        Arguments:
            id {str} -- The ID of the API client to retrieve
        
        Returns:
            ApiClient -- The API client

    Example:        
        .. code-block:: python

            >>> api_client = client.api_clients.get('54759eb3c090d83494e2d804')
            >>> api_client.id
            '54759eb3c090d83494e2d804'
            >>> api_client.role
            'Hybrid'
            >>> api_client.dict()
            {
                'id'='54759eb3c090d83494e2d804',
                'private': False,
                'canRead': [],
                'canWrite': [],
                'owner': '7354308922443094682',
                'createdAt': '2019-06-09T20:23:22+00:00',
                'updatedAt': '2019-06-09T22:38:35+00:00',
                'role': 'Hybrid',
                'documentName': 'Test',
                'documentType': 'DataServer',
                'documentLocation': 'http://some.server.com',
                'online': True,
            }
        """
        return self.make_request('get', '/' + id)

    def update(self, id, data):
        """Update a specific API client
        
        Arguments:
            id {str} -- The ID of the API client to update
            data {dict} -- A dict of values to update
        
        Returns:
            dict -- a confirmation payload with the updated keys

    Example:        
        .. code-block:: python

            >>> api_client.id
            '54759eb3c090d83494e2d804'
            >>> client.api_clients.update(id=api_client.id, data={'documentName': 'newTest', 'documentLocation': 'https://some.new.server.com/dump'})
            {
                "success": True,
                "message": "Client updated following fields: ['documentName', 'documentLocation']"
            }
        """
        return self.make_request('update', '/' + id, data)

    def delete(self, id):
        """Delete a specific API client
        
        Arguments:
            id {str} -- The ID of the API client to delete
        
        Returns:
            dict -- A confirmation payload

    Example:
        .. code-block:: python

            >>> api_client.id
            '54759eb3c090d83494e2d804'
            >>> client.api_clients.delete(id=api_client.id)
            {
                "success": True,
                "message": 'Client was deleted! Bye bye data.'"
            }
        """
        return self.make_request('delete', '/' + id)



import uuid
import pytest
from speckle.SpeckleClient import SpeckleApiClient

@pytest.fixture(scope='session')
def host():
    return 'localhost:3000'

@pytest.fixture(scope='session')
def transfer_protocol():
    return 'http'

@pytest.fixture(scope='session')
def admin_account():
    return {
        'name': 'Testy',
        'surname': 'McTestyFace',
        'company': 'Test Inc',
        'email': 'test@test.com',
        'password': 'supersecretpassword'
    }

@pytest.fixture(scope='session')
def client(host, transfer_protocol, admin_account):
    client = SpeckleApiClient(host=host, transfer_protocol=transfer_protocol)
    try:
        client.register(**admin_account)
    except AssertionError as e:
        if str(e) == 'Email taken. Please login. Thanks!':
            client.login(email=admin_account['email'], password=admin_account['password'])
        else:
            raise e
    
    return client


@pytest.fixture(scope='module')
def user_account(host, transfer_protocol, admin_account):
    client = SpeckleApiClient(
            host=host, transfer_protocol=transfer_protocol)

    account = {
        'name': 'Test_1',
        'surname': 'McTestyFace_1',
        'company': 'Test Inc_1',
        # Can't delete users on TearDown so have to create new on each time for local development
        'email': '{}@test_1.com'.format(str(uuid.uuid4())),
        'password': 'supersecretpassword'
    }

    client.register(**account)
    user = client.me

    return user



"""Speckle Client documentation

The SpeckleClient class is used to manage authentication and dispatch
request to a given SpeckleServer.

Example:
    Instantiate a client to a given server and register/authenticate to it::

        from speckle import SpeckleApiClient

        client = SpeckleApiClient('myspeckle.speckle.works')

        client.register(
            email='test@test.com',
            password='Speckle<3Python',
            company='SnakeySnakes',
            name='Snakey',
            surname='Snake'
        )

        streams = client.streams.list()

For more detailed documentation on the inidividual resources available go  :doc:`here </client_api_ref>`

"""

import requests
from speckle import resources


class ClientBase():
    """Base class for http speckle client

    This class contains the basic properties required to register, authenticate and hold authentication
    credentials.

    """
    
    DEFAULT_HOST = 'hestia.speckle.works'
    DEFAULT_VERSION = 'v1'
    TRANSFER_PROTOCOL = 'https'


    def __init__(self, host=DEFAULT_HOST, version=DEFAULT_VERSION, transfer_protocol=TRANSFER_PROTOCOL):
        self.server = '{}://{}/api/{}'.format(transfer_protocol, host, version)
        self.me = None
        self.s = requests.Session()
        self.verbose = False


    def register(self, email, password, company, name=None, surname=None):
        """Register a new user to the speckle server

        After a user has registered, this function will log them in and save auth token credentials
        in the client object.
        
        Arguments:
            email {str} -- Email address of the new user, must not exist on the server
            password {str} -- Pretty self explanatory, must be at least 8 characters long
            company {str} -- Company the user is registering to speckle under
        
        Keyword Arguments:
            name {str} -- User name, server will default to 'Anonymous' if None (default: {None})
            surname {str} -- User surname, server will default to '' if None (default: {None})
        """

        r = self.s.post(
            self.server + '/accounts/register',
            json={
                'name': name,
                'surname': surname,
                'email': email,
                'password': password,
                'company': company
            })

        response = r.json()
        assert response['success'], response['message']

        self.login(email, password)


    def login(self, email, password):
        """Login user to speckle server
        
        After a user has logged in, this function will save auth token credentials
        in the client object.

        Arguments:
            email {str} -- Email address of the new user, must not exist on the server
            password {str} -- Pretty self explanatory, must be at least 8 characters long. Oh, must also be the user's actual password...
        """

        r = self.s.post(
            self.server + '/accounts/login',
            json={
                'email': email,
                'password': password
            })

        response = r.json()
        assert response['success'], response['message']
        
        self.me = r.json()['resource']
        self.s.headers.update({
            'content-type': 'application/json',
            'Authorization': self.me['token'],
        })

    

    def __getattr__(self, name):
        try:
            attr = getattr(resources, name)
            # setattr(self, name, attr.Resource(self.s, self.server))
            return attr.Resource(self.s, self.server, self.me)
            # return getattr(self, name)
        except:
            raise 'Method {} is not supported by SpeckleClient class'.format(name)

    



import json
from requests import Request
from pydantic import BaseModel
from typing import List, Optional
import dataclasses
from dataclasses import dataclass
from datetime import datetime

class ResourceBaseSchema(BaseModel):
    id: Optional[str]
    private: Optional[bool]
    canRead: Optional[List[str]]
    canWrite: Optional[List[str]]
    owner: Optional[str]
    anonymousComments: Optional[bool]
    comments: Optional[List[str]]
    createdAt: Optional[str]
    updatedAt: Optional[str]

    class Config:
        fields = {'id': '_id'}

class ResourceInfo(BaseModel):
    resourceType: str
    resourceId: str

class Comment(BaseModel):
    id: Optional[str]
    owner: Optional[str]
    comments: Optional[List[str]]
    text: Optional[str]
    flagged: Optional[bool]
    resource: Optional[ResourceInfo]
    otherResources: Optional[List[ResourceInfo]]
    closed: Optional[bool]
    assignedTo: Optional[List[str]]
    labels: Optional[List[str]]
    priority: Optional[str]
    status: Optional[str]
    view: Optional[dict]
    screenshot: Optional[str]

    class Config:
        fields = {'id': '_id'}


def clean_empty(d):
    if not isinstance(d, (dict, list)):
        return d
    if isinstance(d, list):
        return [v for v in (clean_empty(v) for v in d) if v]
    return {k: v for k, v in ((k, clean_empty(v)) for k, v in d.items()) if v}

class ResourceBase(object):

    def __init__(self, session, basepath, me, name, methods):
        self.s = session
        self._path = basepath + '/' + name
        self.me = me
        self._comment_path = basepath + '/comments/' + name
        self.name = name
        self.methods = methods
        self.method_dict = {
            'list': {
                'method': 'GET',
                },
            'create': {
                'method': 'POST',
                },
            'get': {
                'method': 'GET',
                },
            'update': {
                'method': 'PUT',
                },
            'delete': {
                'method': 'DELETE',
                },
            'comment_get': {
                'method': 'GET',
                },
            
            'comment_create': {
                'method': 'POST'}
        }

        for method in self.methods:
            if method == 'retrieve':
                self.__setattr__('retrieve', 'x')

        self.schema = None
        self.comment_schema = Comment

    def _prep_request(self, method, path, comment, data):
        assert method in self.methods, 'method {} not supported for {} calls'.format(method, self.name)

        if comment:
            url = self._comment_path + path
            if data:
                dataclass_instance = self.comment_schema.parse_obj(data)
                data = clean_empty(dataclass_instance.dict()) 
        else:
            url = self._path + path
            if data:
                if self.schema:
                    dataclass_instance = self.schema.parse_obj(data)
                    data = clean_empty(dataclass_instance.dict())

        return self.s.prepare_request(Request(self.method_dict[method]['method'], url, json=data))

    def _parse_response(self, response, comment=False, schema=None):
        if schema:
            return schema.parse_obj(response)
        elif comment:
            return self.comment_schema.parse_obj(response)
        elif self.schema:
            return self.schema.parse_obj(response)
        else:
            return response


    def make_request(self, method, path, data=None, comment=False, schema=None):
        r = self._prep_request(method, path, comment, data)
        resp = self.s.send(r)

        response_payload = resp.json()
        print(json.dumps(response_payload, indent=2))

        assert response_payload['success'] == True, json.dumps(response_payload)
        # print(response_payload)
        if 'resources' in response_payload:
            return [self._parse_response(resource, comment, schema) for resource in response_payload['resources']]
        elif 'resource' in response_payload:
            return self._parse_response(response_payload['resource'], comment, schema)
        else:
            return response_payload # Not sure what to do in this scenario or when it might occur



import asyncio
import websockets


async def hello():
    async with websockets.connect(
            'wss://hestia.speckle.works/') as websocket:
        name = input("What's your name? ")

        await websocket.send(name)
        print(f"> {name}")

        greeting = await websocket.recv()
        print(f"< {greeting}")

asyncio.get_event_loop().run_until_complete(hello())

# websockets.client.


from speckle.base.resource import ResourceBase, ResourceBaseSchema
from typing import List, Optional


NAME = 'comments'
METHODS = ['list', 'get', 'update', 'assigned',
           'delete', 'comment_get', 'comment_create']


class Resource(ResourceBase):
    """API Access class for Comments

    """

    def __init__(self, session, basepath, me):
        super().__init__(session, basepath, me, NAME, METHODS)

        self.schema = self.comment_schema

        self.method_dict.update({
            'assigned': {
                    'method': 'GET'
                }
        })

    def list(self):
        """List all comments
        
        Returns:
            list -- A list of comment data class instances
        """
        return self.make_request('list', '/')


    def get(self, id):
        """Get a specific comment from the SpeckleServer
        
        Arguments:
            id {str} -- The ID of the comment to retrieve
        
        Returns:
            Comment -- The comment
        """
        return self.make_request('get', '/' + id)

    def update(self, id, data):
        """Update a specific comment
        
        Arguments:
            id {str} -- The ID of the comment to update
            data {dict} -- A dict of values to update
        
        Returns:
            dict -- a confirmation payload with the updated keys
        """
        return self.make_request('update', '/' + id, data)

    def delete(self, id):
        """Delete a specific comment
        
        Arguments:
            id {str} -- The ID of the comment to delete
        
        Returns:
            dict -- A confirmation payload
        """
        return self.make_request('delete', '/' + id)

    def comment_get(self, id):
        """Retrieve comments attached to a comment
        
        Arguments:
            id {str} -- The ID of the comment to retrieve comments from
        
        Returns:
            list -- A list of comments
        """
        return self.make_request('comment_get', '/' + id, comment=True)

    def comment_create(self, id, data):
        """Add a comment to a comment
        
        Arguments:
            id {str} -- The ID of the comment to comment on
            data {dict} -- A comment dictionary object
        
        Returns:
            Comment -- The comment created by the server
        """
        return self.make_request('comment_create', '/' + id, data, comment=True)

    def assigned(self):
        """Get the list of all comments where the logged in user is assigned
        
        Returns:
            list -- A list of comments the user is assigned to
        """
        return self.make_request('assigned', '/assigned')



import pytest


@pytest.fixture(scope='module')
def resource():
    return {
        'text': 'Wow this is an amazing comment!',
    }

@pytest.fixture
def comment():
    return {
        'text': 'Comments within comments... woah!',
    }


@pytest.fixture(scope='module')
def created_item(client, resource):
    project = client.projects.create({
        'name': 'test project',
        'description': 'a project made for testing purposes',
        'tags': [],
        'streams': [],
    })
    yield client.projects.comment_create(project.id, resource)
    client.projects.delete(project.id)

@pytest.mark.dependency()
def test_list(client, resource, created_item):
    # By this stage we should have created a bunch of comments
    initial_count = client.comments.list()
    assert len(initial_count) >= 1


@pytest.mark.dependency()
def test_get(client, created_item):
    returned_resource = client.comments.get(id=created_item.id)

    assert created_item.id == returned_resource.id
    assert created_item.text == returned_resource.text

@pytest.mark.dependency(depends=['test_get'])
def test_comments(client, created_item, comment):
    comments = client.comments.comment_get(created_item.id)
    assert comments == []
    client.comments.comment_create(created_item.id, comment)
    comments = client.comments.comment_get(created_item.id)
    assert len(comments) == 1
    assert comments[0].text == comment['text']


@pytest.mark.dependency(depends=['test_get'])
def test_assigned(client, created_item):
    assigned = client.comments.assigned()
    assert assigned == []
    created_item.assignedTo = [client.me['_id']]
    client.comments.update(created_item.id, created_item.dict())
    assigned = client.comments.assigned()
    assert len(assigned) == 1
    assert assigned[0].id == created_item.id

@pytest.mark.dependency(depends=['test_comments'])
def test_delete(client, resource):
    data = client.comments.list()
    assert data != []
    for comment in data:
        client.comments.delete(id=comment.id)



import uuid
import pytest


@pytest.fixture(scope='module')
def resource():
    return {
        'name': 'test project',
        'description': 'a project made for testing purposes',
        'tags': [],
        'streams': [],
    }


@pytest.fixture(scope='module')
def updated_resource():
    return {
        'name': 'test project',
        'description': 'a project made for testing purposes',
        'tags': [],
        'streams': [],
    }


@pytest.fixture(scope='module')
def streams(client):
    stream_1 = client.streams.create({
        'name': 'example stream 1',
        })
    stream_2 = client.streams.create({
        'name': 'example stream 2',
        })
    yield [stream_1, stream_2]
    streams = client.streams.list()
    for stream in streams:
        client.streams.delete(id=stream.streamId)


@pytest.fixture(scope='module')
def created_item(client, resource):
    return client.projects.create(resource)
    

@pytest.fixture
def comment():
    return {
        'text': 'Wow this is an amazing resource!',
    }

def test_create(client, resource):
    created_item = client.projects.create(resource)

    assert resource['name'] == created_item.name
    assert resource['description'] == created_item.description
    assert resource['tags'] == created_item.tags
    assert resource['streams'] == created_item.streams

    assert created_item.id is not None


@pytest.mark.dependency(depends=['test_create'])
def test_list(client, resource):
    assert len(client.projects.list()) == 1
    client.projects.create(resource)
    assert len(client.projects.list()) == 2


@pytest.mark.dependency(depends=['test_create'])
def test_get(client, created_item):
    returned_resource = client.projects.get(id=created_item.id)

    assert created_item.id == returned_resource.id
    assert created_item.name == returned_resource.name
    assert created_item.description == returned_resource.description
    assert created_item.tags == returned_resource.tags
    assert created_item.streams == returned_resource.streams

@pytest.mark.dependency(depends=['test_get'])
def test_add_stream(client, created_item, streams):
    created_item.streams = []

    client.projects.add_stream(created_item.id, streams[0].streamId)
    project = client.projects.get(created_item.id)
    assert project.streams == [streams[0].streamId]

    client.projects.add_stream(created_item.id, streams[1].streamId)
    project = client.projects.get(created_item.id)
    assert project.streams == [streams[0].streamId, streams[1].streamId]


@pytest.mark.dependency(depends=['test_add_stream'])
def test_remove_stream(client, created_item, streams):
    project = client.projects.get(created_item.id)
    assert len(project.streams) == 2

    client.projects.remove_stream(project.id, streams[1].streamId)
    project = client.projects.get(created_item.id)

    assert project.streams == [streams[0].streamId]

@pytest.mark.dependency(depends=['test_create'])
def test_user_manipulation(client, created_item, user_account):
    project = client.projects.get(created_item.id)

    assert project.permissions.canRead == []
    assert project.permissions.canWrite == []

    client.projects.add_user(created_item.id, user_account['_id'])
    project = client.projects.get(created_item.id)

    assert project.canRead == [user_account['_id']]
    assert project.permissions.canWrite == [user_account['_id']]

    client.projects.downgrade_user(created_item.id, user_account['_id'])
    project = client.projects.get(created_item.id)

    assert project.canRead == [user_account['_id']]
    assert project.permissions.canWrite == []

    client.projects.upgrade_user(created_item.id, user_account['_id'])
    project = client.projects.get(created_item.id)

    assert project.canRead == [user_account['_id']]
    assert project.permissions.canWrite == [user_account['_id']]

    client.projects.remove_user(created_item.id, user_account['_id'])
    project = client.projects.get(created_item.id)

    assert project.canRead == []
    assert project.permissions.canWrite == []


@pytest.mark.dependency(depends=['test_create'])
def test_comments(client, created_item, comment):
    comments = client.projects.comment_get(created_item.id)
    assert comments == []
    client.projects.comment_create(created_item.id, comment)
    comments = client.projects.comment_get(created_item.id)
    assert len(comments) == 1
    assert comments[0].text == comment['text']


@pytest.mark.dependency(depends=['test_comments'])
def test_delete(client, resource):
    data = client.projects.list()
    assert data != []
    for project in data:
        client.projects.delete(id=project.id)



from speckle.base.resource import ResourceBase
from pathlib import Path
import sys
import inspect
import pkgutil
from importlib import import_module


for (_, name, _) in pkgutil.iter_modules([Path(__file__).parent]):

    imported_module = import_module('.' + name, package=__name__)
    
    if hasattr(imported_module, 'Resource'):
        setattr(sys.modules[__name__], name, imported_module)


from speckle.base.resource import ResourceBase
from pydantic import BaseModel
from typing import List, Optional
from speckle.base.resource import ResourceBase, ResourceBaseSchema

NAME = 'projects'
METHODS = [
    'list', 'create', 'get', 'update', 'delete',
    'comment_get', 'comment_create', 'add_stream',
    'add_user', 'remove_user', 'remove_stream',
    'upgrade_user', 'downgrade_user'
    ]

class Permissions(BaseModel):
    canRead: Optional[List[str]]
    canWrite: Optional[List[str]]

class Project(ResourceBaseSchema):
    """Project Data Class
    
    Arguments:
        ResourceBaseSchema {BaseModel} -- The BaseModel for all Speckle objects
    """
    name: Optional[str]
    description: Optional[str]
    tags: List[str] = []
    streams: List[str] = []
    permissions: Optional[Permissions]


class Resource(ResourceBase):
    """API Access class for Projects

    """
    def __init__(self, session, basepath, me):
        super().__init__(session, basepath, me, NAME, METHODS)

        self.method_dict.update({
            'add_stream': {
                'method': 'PUT'
            },
            'remove_stream': {
                'method': 'DELETE'
            },
            'add_user': {
                'method': 'PUT'
            },
            'remove_user': {
                'method': 'DELETE'
            },
            'upgrade_user': {
                'method': 'PUT'
            },
            'downgrade_user': {
                'method': 'PUT'
            },
        })

        self.schema = Project

    def list(self):
        """List all projects
        
        Returns:
            list -- A list of project data class instances
        """
        return self.make_request('list', '/')

    def create(self, data):
        """Create a project from a data dictionary
        
        Arguments:
            data {dict} -- A dictionary describing a project
        
        Returns:
            Project -- The instance created on the Speckle Server
        """
        return self.make_request('create', '/', data)

    def get(self, id):
        """Get a specific project from the SpeckleServer
        
        Arguments:
            id {str} -- The ID of the project to retrieve
        
        Returns:
            Project -- The project
        """
        return self.make_request('get', '/' + id)

    def update(self, id, data):
        """Update a specific project
        
        Arguments:
            id {str} -- The ID of the project to update
            data {dict} -- A dict of values to update
        
        Returns:
            dict -- a confirmation payload with the updated keys
        """
        return self.make_request('update', '/' + id, data)

    def delete(self, id):
        """Delete a specific project
        
        Arguments:
            id {str} -- The ID of the project to delete
        
        Returns:
            dict -- A confirmation payload
        """
        return self.make_request('delete', '/' + id)

    def comment_get(self, id):
        """Retrieve comments attached to a project
        
        Arguments:
            id {str} -- The ID of the project to retrieve comments from
        
        Returns:
            list -- A list of comments
        """
        return self.make_request('comment_get', '/' + id, comment=True)

    def comment_create(self, id, data):
        """Add a comment to a project
        
        Arguments:
            id {str} -- The ID of the project to comment on
            data {dict} -- A comment dictionary object
        
        Returns:
            CommentSchema -- The comment created by the server
        """
        return self.make_request('comment_create', '/' + id, data, comment=True)


    def add_stream(self, id, stream_id):
        """Add a Stream to a project
        
        Arguments:
            id {str} -- The ID of a project
            stream_id {str} -- The StreamId of a stream
        
        Returns:
            dict -- A dictionary with the stream and the project
        """
        return self.make_request('add_stream', '/' + id + '/addstream/' + stream_id)

    def remove_stream(self, id, stream_id):
        """Remove a stream from a project
        
        Arguments:
            id {str} -- The ID of a project
            stream_id {str} -- The StreamId of a stream
        
        Returns:
            dict -- A dictionary with the stream and the project
        """
        return self.make_request('remove_stream', '/' + id + '/removestream/' + stream_id)
    
    def add_user(self, id, user_id):
        """Add a user to a project.

        Note:
            When a user is first added to a project they have read and write
            authorizations. If you only want to allow them to write be sure to
            downgrade the user right after adding them.
        
        Arguments:
            id {str} -- The ID of a project
            user_id {str} -- The ID of a user
        
        Returns:
            dict -- A confirmation payload
        """
        return self.make_request('add_user', '/' + id + '/adduser/' + user_id)

    def remove_user(self, id, user_id):
        """Remove a user from a project

        Note:
            When you remove a user from a project this also removes all of their
            read and write access to the project's streams.
        
        Arguments:
            id {str} -- The ID of a project
            user_id {str} -- The ID of a user
        
        Returns:
            dict -- A confirmation payload
        """
        return self.make_request('remove_user', '/' + id + '/removeuser/' + user_id)

    def upgrade_user(self, id, user_id):
        """Upgrade a user to have Write access to the project and it's streams
        
        Arguments:
            id {str} -- The ID of a project
            user_id {str} -- The ID of a user
        
        Returns:
            dict -- A confirmation payload
        """
        return self.make_request('upgrade_user', '/' + id + '/upgradeuser/' + user_id)

    def downgrade_user(self, id, user_id):
        """Downgrade a user to have only Read access to the project and it's streams
        
        Arguments:
            id {str} -- The ID of a project
            user_id {str} -- The ID of a user
        
        Returns:
            dict -- A confirmation payload
        """
        return self.make_request('downgrade_user', '/' + id + '/downgradeuser/' + user_id)



import uuid
import pytest


@pytest.fixture(scope='module')
def resource():
    return {
        'role': 'Hybrid',
        'documentName': 'Test',
        'documentType': 'DataServer',
        'documentLocation': 'http://some.server.com',
        'documentGuid': 'some-long-uuid',
        'online': True,
    }


@pytest.fixture(scope='module')
def updated_resource():
    return {
        'documentName': 'Updated Test',
        'documentLocation': 'http://some.other.server.com',
    }

@pytest.fixture(scope='module')
def stream(client):
    stream = client.streams.create({
        'name': 'example stream',
        'description': 'a stream for testing purposes',
        'tags': [
            'test',
            'example'
        ],
        'commitMessage': 'created stream',
        'objects': [],
        'layers': [],
    })
    yield stream
    client.streams.delete(id=stream.streamId)

def test_list_none(client):
    data = client.api_clients.list()
    assert data == []


def test_create(client, resource, stream):
    resource['streamId'] = stream.streamId

    returned_resource = client.api_clients.create(resource)

    assert resource['role'] == returned_resource.role
    assert resource['documentName'] == returned_resource.documentName
    assert resource['documentType'] == returned_resource.documentType
    assert resource['documentLocation'] == returned_resource.documentLocation
    assert resource['documentGuid'] == returned_resource.documentGuid
    assert resource['streamId'] == returned_resource.streamId
    assert resource['online'] == returned_resource.online
    

@pytest.mark.dependency(depends=['test_create'])
def test_get(client, resource):
    data = client.api_clients.list()
    assert len(data) == 1
    returned_resource = client.api_clients.get(id=data[0].id)

    assert resource['role'] == returned_resource.role
    assert resource['documentName'] == returned_resource.documentName
    assert resource['documentType'] == returned_resource.documentType
    assert resource['documentLocation'] == returned_resource.documentLocation
    assert resource['documentGuid'] == returned_resource.documentGuid
    assert resource['online'] == returned_resource.online

@pytest.mark.dependency(depends=['test_get'])
def test_update(client, updated_resource):
    data = client.api_clients.list()
    assert len(data) == 1
    response = client.api_clients.update(id=data[0].id, data=updated_resource)

    assert response['message'] == 'Client updated following fields: documentName,documentLocation,documentGuid'

    returned_resource = client.api_clients.get(id=data[0].id)

    assert updated_resource['documentName'] == returned_resource.documentName
    assert updated_resource['documentLocation'] == returned_resource.documentLocation

@pytest.mark.dependency(depends=['test_create'])
def test_delete(client):
    data = client.api_clients.list()
    assert data != []
    for api_client in data:
        client.api_clients.delete(id=api_client.id)
    data = client.api_clients.list()
    assert data == []



import uuid
import pytest
from speckle.SpeckleClient import SpeckleApiClient


# EMAIL_UUID = str(uuid.uuid4())

@pytest.fixture(scope='module')
def account():
    return {
        'name': 'Test_1',
        'surname': 'McTestyFace_1',
        'company': 'Test Inc_1',
        # Can't delete users on TearDown so have to create new on each time for local development
        'email': '{}@test_1.com'.format(str(uuid.uuid4())),
        'password': 'supersecretpassword'
    }


@pytest.fixture()
def new_profile_payload():
    return {
        'name': 'name_test',
        'surname': 'surname_test',
        'company': 'company_test',
        'role': 'not-a-user'
    }


@pytest.fixture(scope='module')
def created_item(host, transfer_protocol):
    client = SpeckleApiClient(
            host=host, transfer_protocol=transfer_protocol)
    account = {
        'name': 'Test_1',
        'surname': 'McTestyFace_1',
        'company': 'Test Inc_1',
        # Can't delete users on TearDown so have to create new on each time for local development
        'email': '{}@test_1.com'.format(str(uuid.uuid4())),
        'password': 'supersecretpassword'
    }
    client.register(**account)
    return client.me

@pytest.fixture
def get_user(host, transfer_protocol, created_item):
    def get_me():
        client = SpeckleApiClient(
            host=host, transfer_protocol=transfer_protocol)
        client.login(email=created_item['email'], password='supersecretpassword')
        return client.me
    return get_me


def test_register(host, transfer_protocol, account):
    test_client = SpeckleApiClient(host=host, transfer_protocol=transfer_protocol)
    assert test_client.me == None
    test_client.register(**account)
    assert test_client.me['email'] == account['email']
    assert test_client.me['apitoken'] is not None
    assert test_client.me['token'] is not None


@pytest.mark.dependency(depends=['test_register'])
def test_login(host, transfer_protocol, account):
    client = SpeckleApiClient(host=host, transfer_protocol=transfer_protocol)
    assert client.me == None
    client.login(email=account['email'], password=account['password'])
    assert client.me['email'] == account['email']
    assert client.me['apitoken'] is not None
    assert client.me['token'] is not None


@pytest.mark.dependency(depends=['test_login'])
def test_list(host, transfer_protocol, account):
    client = SpeckleApiClient(host=host, transfer_protocol=transfer_protocol)
    client.login(email=account['email'], password=account['password'])
    try:
        client.accounts.list()
    except AttributeError as e:
        assert str(e) == "'Resource' object has no attribute 'list'"


@pytest.mark.dependency(depends=['test_login'])
def test_create(host, transfer_protocol, account):
    client = SpeckleApiClient(host=host, transfer_protocol=transfer_protocol)
    client.login(email=account['email'], password=account['password'])
    try:
        client.accounts.create(data=account)
    except AttributeError as e:
        assert str(e) == "'Resource' object has no attribute 'create'"


@pytest.mark.dependency(depends=['test_login'])
def test_delete(host, transfer_protocol, account):
    client = SpeckleApiClient(host=host, transfer_protocol=transfer_protocol)
    client.login(email=account['email'], password=account['password'])
    try:
        client.accounts.delete(id='some-made-up-id')
    except AttributeError as e:
        assert str(e) == "'Resource' object has no attribute 'delete'"


def test_get(client, created_item):
    account = client.accounts.get(created_item['_id'])

    assert account['_id'] == created_item['_id']


@pytest.mark.dependency(depends=['test_login'])
def test_get_profile(host, transfer_protocol, account):
    client = SpeckleApiClient(host=host, transfer_protocol=transfer_protocol)
    assert client.me == None
    client.login(email=account['email'], password=account['password'])
    profile = client.accounts.get_profile()

    assert profile['email'] == account['email']
    assert profile['name'] == account['name']
    assert profile['surname'] == account['surname']
    assert profile['company'] == account['company']
    assert profile['private'] == True
    assert profile['role'] == 'admin'
    assert profile['_id'] is not None


@pytest.mark.dependency(depends=['test_get_profile'])
def test_update_profile(host, transfer_protocol, account, new_profile_payload):
    client = SpeckleApiClient(host=host, transfer_protocol=transfer_protocol)
    assert client.me == None
    client.login(email=account['email'], password=account['password'])
    
    client.accounts.update_profile(new_profile_payload)
    profile = client.accounts.get_profile()

    assert profile['name'] != account['name']
    assert profile['surname'] != account['surname']
    assert profile['company'] != account['company']

    assert profile['name'] == new_profile_payload['name']
    assert profile['surname'] == new_profile_payload['surname']
    assert profile['company'] == new_profile_payload['company']

    assert profile['role'] == 'admin'
    assert profile['role'] != new_profile_payload['role']


@pytest.mark.dependency(depends=['test_get'])
def test_set_role(client, get_user):
    # account = get_user()
    # assert client.accounts.get_profile() == {}
    # assert account['role'] == 'user'

    # client.accounts.set_role(account['_id'], 'admin')
    # account = get_user()
    # assert account['role'] == 'admin'

    # client.accounts.set_role(account['_id'], 'user')
    # account = get_user()
    # assert account['role'] == 'user'
    # TODO(): Fix this mess...
    pass





import requests, json, os
import sqlite3, contextlib
import struct, base64
from speckle.base.client import ClientBase

def jdumps(msg):
    return json.dumps(msg, indent=4, sort_keys=True)

class SpeckleApiClient(ClientBase):

    '''
    Utility functions
    '''

    def log(self, msg):
        print ('SpeckleClient: {}'.format(msg))

    def check_response_status_code(self, r):

        # Parse response
        if r.status_code == 200:
            self.log("Request successful: {}".format(r.reason))
            return True
        elif r.status_code == 400:
            self.log("Request failed: {}".format(r.reason))
        elif r.status_code != 200 and r.status_code != 204:
            self.log("The HTTP status code of the response was not expected: {},  {}".format(r.status_code, r.reason))
        else:
            self.log("Unknown response: {},  {}".format(r.status_code, r.reason))            
        # Debug
        if self.verbose:
            print()
            self.log(json.dumps(r.text, indent=4, sort_keys=True))            
        return False

    def write_profile_to_file(self, profile, directory=None):
        assert ("email" in profile.keys())
        assert ("server" in profile.keys())
        assert ("server_name" in profile.keys())
        assert ("apitoken" in profile.keys())

        if directory == None:
            directory = os.path.join(os.getenv('LOCALAPPDATA'), os.path.join('SpeckleSettings', 'MigratedAccounts'))

        if os.path.isdir(directory):
            with open(os.path.join(directory, profile['email']) + ".JWT .txt", 'w') as f:
                f.write("{},{},{},{},{}\n".format(profile['email'], profile['apitoken'], profile['server_name'], profile['server'], profile['server']))
    
    def write_profile_to_database(self, profile, filepath=None):
        assert ("email" in profile.keys())
        assert ("server" in profile.keys())
        assert ("server_name" in profile.keys())
        assert ("apitoken" in profile.keys())

        if filepath == None:
            filepath = os.path.join(os.getenv('LOCALAPPDATA'), os.path.join('SpeckleSettings', 'SpeckleCache.db'))

        if os.path.isfile(filepath):
            try:
                conn = sqlite3.connect(filepath)
            except:
                print ("Accessing database failed...")
                return None

            #with contextlib.closing(sqlite3.connect(filepath)) as con:
            with conn:
                c = conn.cursor()
                c.execute(""" INSERT INTO Account(AccountId,ServerName,RestApi,Email,Token,IsDefault)
                             VALUES(NULL,?,?,?,?,0) """, (profile['server_name'], profile['server'], profile['email'], profile['apitoken']))
                conn.commit()
                return c.lastrowid

    def load_local_profiles_from_database(self, filepath=None):
        profiles = []
        if filepath == None:
            filepath = os.path.join(os.getenv('LOCALAPPDATA'), os.path.join('SpeckleSettings', 'SpeckleCache.db'))

        accounts = None
        if os.path.isfile(filepath):
            with contextlib.closing(sqlite3.connect(filepath)) as con:
                c = con.cursor()
                c.execute("SELECT * FROM Account")
                accounts = c.fetchall()
                #names = [x[0] for x in c.description]
                for a in accounts:
                    profiles.append({'server_name':a[1], 'server':a[2], 'email':a[3], 'apitoken':a[4]})
        return profiles


    def load_local_profiles(self, directory=None):
        profiles = []

        if directory == None:
            directory = os.path.join(os.getenv('LOCALAPPDATA'), os.path.join('SpeckleSettings', 'MigratedAccounts'))

        if os.path.isdir(directory):
            files = [os.path.join(directory, x) for x in os.listdir(directory) if x.endswith('.txt')]
            for file in files:
                with open(file, 'r') as f:
                    line = f.readline()
                    tokens = line.split(',')

                    if len(tokens) < 4: return None

                    profiles.append({'email':tokens[0], 'server_name': tokens[2], 'apitoken': tokens[1], 'server': tokens[3]})

        return profiles

    def set_profile(self, server, apitoken):
        self.server = server
        self.s.headers.update({'Authorization': apitoken})

    def use_existing_profile(self, email, directory=None):
        profiles = self.load_local_profiles(directory)
        profile = profiles[0] # Maybe worth implementing a retrieve by id?
        if email in profile.keys():
            self.server = profiles[email]['server']
            self.s.headers.update({'Authorization': profiles[email]['apitoken']})
            return True
        return False

    '''
    API calls
    '''

    def ClientCreateAsync(self, client):
        '''
        Create client
        '''
        url = self.server + "/clients"
        r = self.s.post(url, json.dumps(client))

        if self.check_response_status_code(r):
            return r.json()
        return None

    def ClientDeleteAsync(self, client):
        raise NotImplementedError

    def ClientGetAllAsync(self, query=""):
        '''
        Gets a user's profile
        '''
        url = self.server + "/clients?{}".format(query)
        r = self.s.get(url)

        if self.check_response_status_code(r):
            return r.json()
        return None

    def ClientGetAsync(self, client):
        raise NotImplementedError

    def ClientUpdateAsync(self, clientId, client):
        '''
        Update client
        '''
        url = self.server + "/clients/{}".format(clientId)
        r = self.s.put(url, json.dumps(client))

        if self.check_response_status_code(r):
            return r.json()
        return None

    def CommentCreateAsync(self, resourceType, str, comment):
        raise NotImplementedError

    def CommentDeleteAsync(self, str):
        raise NotImplementedError

    def CommentGetAsync(self, str):
        raise NotImplementedError

    def CommentGetFromResourceAsync(self, resourceType, str):
        raise NotImplementedError

    def CommentUpdateAsync(self, str, comment):
        raise NotImplementedError
    '''
    def GetObjectData(System.Runtime.Serialization.SerializationInfo, System.Runtime.Serialization.StreamingContext):
        raise NotImplementedError
    
    def JoinRoom(self, str):
        raise NotImplementedError

    def LeaveRoom(self, str):
        raise NotImplementedError
    '''

    def ObjectCreateAsync(self, objectList):
        '''
        Create objects
        '''
        payload = []
        for o in objectList:
            if '_id' in o.keys():
                del o['_id']
            if 'hash' in o.keys():
                del o['hash']

            payload.append(o)

        url = self.server + "/objects"
        r = self.s.post(url, json.dumps(payload))

        if self.check_response_status_code(r):
            return r.json()
        return None

    def ObjectDeleteAsync(self, objectId):
        '''
        Delete a specific object
        '''
        url = self.server + "/objects/{}".format(objectId)
        r = self.s.delete(url)

        if self.check_response_status_code(r):
            return r.json()
        return None

    def ObjectGetAsync(self, objectId, query=""):
        '''
        Get a specific object
        '''
        url = self.server + "/objects/{}?{}".format(objectId, query)
        r = self.s.get(url)

        if self.check_response_status_code(r):
            return r.json()
        return None

    def ObjectGetBulkAsync(self, objectIds, query=""):
        '''
        Get a list of objects at once
        '''
        url = self.server + "/objects/getbulk?{}".format(query)
        r = self.s.post(url, data=json.dumps(objectIds))

        if self.check_response_status_code(r):
            return r.json()
        return None

    def ObjectUpdateAsync(self, objectId, speckle_object):
        '''
        Update object.
        '''
        assert objectId is not None
        url = self.server + "/objects/{}".format(objectId)
        r = self.s.put(url, json.dumps(speckle_object))

    def ObjectUpdatePropertiesAsync(self, objectId, prop):
        url = self.server + "/objects/{}/properties".format(objectId)
        r = self.s.put(url, json.dumps(prop))

        if self.check_response_status_code(r):
            return r.json()
        return None

    def ProjectCreateAsync(self, project):
        '''
        Create project
        '''
        url = self.server + "/projects"
        r = self.s.post(url, json.dumps(project))

        if self.check_response_status_code(r):
            return r.json()
        return None

    def ProjectDeleteAsync(self, projectId):
        '''
        Delete a project using the projectId
        '''
        url = self.server + "/projects/{}".format(projectId)
        r = self.s.delete(url)

        if self.check_response_status_code(r):
            return r.json()
        return None

    def ProjectGetAllAsync(self, query=""):
        '''
        Get all of a user's projects
        '''
        url = self.server + "/projects"
        if query:
            url += "?{}".format(query)
        r = self.s.get(url)

        if self.check_response_status_code(r):
            return r.json()
        return None

    def ProjectGetAsync(self, projectId):
        '''
        Get project by id
        '''
        url = self.server + "/projects/{}".format(projectId)
        r = self.s.get(url)

        if self.check_response_status_code(r):
            return r.json()
        return None

    def ProjectUpdateAsync(self, projectId, project):
        '''
        Update an existing project
        '''
        url = self.server + "/projects/{}".format(projectId)
        r = self.s.put(url, json.dumps(project))

        if self.check_response_status_code(r):
            return r.json()
        return None

    def StreamCloneAsync(self, streamId):
        '''
        Clone an exsting stream
        '''
        url = self.server + "/streams/{}/clone".format(streamId)
        r = self.s.post(url)

        if self.check_response_status_code(r):
            return r.json()
        return None

    def StreamCreateAsync(self, stream):
        '''
        Create new stream
        '''
        url = self.server + "/streams"
        r = self.s.post(url, data=json.dumps(stream))

        if self.check_response_status_code(r):
            return r.json()
        return None

    def StreamDeleteAsync(self, streamId):
        '''
        Delete stream.
        '''
        assert streamId is not None
        url = self.server + "/streams/{}".format(streamId)
        r = self.s.delete(url)

        if self.check_response_status_code(r):
            return r.json()
        return None

    def StreamDiffAsync(self, streamId1, streamId2):
        '''
        Diff two streams
        '''
        url = self.server + "/streams/{}/diff/{}".format(streamId1, streamId2)
        r = self.s.get(url)

        if self.check_response_status_code(r):
            return r.json()
        return None

    def StreamGetAsync(self, streamId, query=""):
        '''
        Get stream
        '''
        url = self.server + "/streams/{}?{}".format(streamId, query)
        r = self.s.get(url)

        if self.check_response_status_code(r):
            return r.json()
        return None

    def StreamGetObjectsAsync(self, streamId, query=""):
        '''
        Get objects on stream..
        '''
        url = self.server + "/streams/{}/objects?{}".format(streamId, query)
        r = self.s.get(url)

        if self.check_response_status_code(r):
            return r.json()
        return None

    def StreamsGetAllAsync(self, query=""):
        '''
        Get all streams available to user.
        '''
        url = self.server + "/streams?omit=objects"
        r = self.s.get(url)

        if self.check_response_status_code(r):
            return r.json()
        return None

    def StreamUpdateAsync(self, streamId, stream):
        '''
        Update stream.
        '''
        assert streamId is not None
        url = self.server + "/streams/{}".format(streamId)
        r = self.s.put(url, json.dumps(stream))

        if self.check_response_status_code(r):
            return r.json()
        return None

    def UserGetAsync(self):
        url = self.server + "/accounts"
        r = self.s.get(url)

        if self.check_response_status_code(r):
            return r.json()
        return None

    def UserGetProfileByIdAsync(self, userId):
        '''
        Get a user's profile
        '''
        url = self.server + "/accounts/{}".format(userId)
        r = self.s.get(url)

        if self.check_response_status_code(r):
            return r.json()
        return None

    def UserLoginAsync(self, user):
        '''
        Login with email and password. 
        user - dictionary that must contain 'email' and 'password' keys
        '''
        assert ("email" in user.keys())
        assert ("password" in user.keys())

        url = self.server + "/accounts/login"

        if self.verbose:
            print(url)

        r = self.s.post(url, data=json.dumps(user))

        if self.check_response_status_code(r):
            login_response = r.json()
            self.s.headers.update(
                {'Authorization': login_response['resource']['apitoken']})

            return login_response
        return None

    def UserRegisterAsync(self, user):
        assert ("server_name" in user.keys())
        assert ("email" in user.keys())
        assert ("password" in user.keys())
        assert ("name" in user.keys())
        assert ("surname" in user.keys())

        url = self.server + "/accounts/register"
        r = self.s.post(url, data=json.dumps(user))

        if self.check_response_status_code(r):
            return r.json()
        return None

    def UserSearchAsync(self, user):

        raise NotImplementedError

    def UserUpdateProfileAsync(self, user):
        assert ("server_name" in user.keys())
        assert ("email" in user.keys())
        assert ("password" in user.keys())
        assert ("name" in user.keys())
        assert ("surname" in user.keys())

        url = self.server + "/accounts/register"
        r = self.s.put(url, data=json.dumps(user))

        if self.check_response_status_code(r):
            return r.json()
        return None



import re
import setuptools
import sys

with open('requirements.txt') as f:
    required = f.read().splitlines()

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="speckle",
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    author="Speckle Works",
    author_email="devops@speckle.works",
    description="A Python client for Speckle servers.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/specklworks.pyspeckle",
    packages=setuptools.find_packages(exclude=["tests"]),
    install_requires=required,
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Operating System :: OS Independent"
    ],
)



from pydantic import BaseModel
from typing import Optional
from speckle.base.resource import ResourceBase


NAME = 'accounts'
METHODS = ['get', 'get_profile', 'update_profile', 'set_role', 'search']


class User(BaseModel):
    id: Optional[str]
    name: Optional[str]
    surname: Optional[str]
    company: Optional[str]
    avatar: Optional[str]
    role: Optional[str]

    class Config:
        fields = {'id': '_id'}

class Resource(ResourceBase):
    """API Access class for Accounts
    
    The accounts resource is used to search, retrieve and manipulate user profiles. None of the methods return an instanciated data class, instead they all return dictionary payloads.

    Example:
        Here is an example of what an account object looks like: 
        
        .. code-block:: python

            {
                "id": "507f1f77bcf86cd799439011"
                "name": "Test"
                "surname": "McTestyFace"
                "company": "Acme"
                "avatar": "https://encrypted-tbn0.gstatic.com/images?q=tbn:ANd9GcRDzvciiQ5-P6itEjycriGb9l9sJH9E538C-tM9QRgFVTaj0Muq"
                "role": "user"
            }
    
    """
    
    def __init__(self, session, basepath, me):
        super().__init__(session, basepath, me, NAME, METHODS)

        # self.schema = User

        self.method_dict.update({
            'get_profile': {
                'method': 'GET'
            },
            'update_profile': {
                'method': 'PUT'
            },
            'set_role': {
                'method': 'PUT'
            },
            'search': {
                'method': 'POST'
            },
        })


    def get(self, id):
        """Get a specific user from the SpeckleServer

        Arguments:
            id {str} -- The ID of the resource to retrieve
        
        Returns:
            dict -- The user

        Example:
            .. code-block:: python

                >>> user_id = '507f1f77bcf86cd799439011'
                >>> client.accounts.get(id=user_id)
                {
                    "success": True,
                    "resource": {
                        "id": "507f1f77bcf86cd799439011",
                        "email": "test@test.com"
                        "name": "Test"
                        "surname": "McTestyFace"
                        "company": "Acme"
                        "avatar": "https://encrypted-tbn0.gstatic.com/images?q=tbn:ANd9GcRDzvciiQ5-P6itEjycriGb9l9sJH9E538C-tM9QRgFVTaj0Muq"
                        "role": "user"
                    }
                }
        """
        return self.make_request('get', '/' + id)

    def get_profile(self):
        """Get current logged in user's profile
        
        Returns:
            dict -- The current user's profile

        Example:
            .. code-block:: python

                >>> from speckle.SpeckleClient import SpeckleApiClient
                >>> client = SpeckleApiClient(host='hestia.speckle.com')
                >>> client.login(email='test@test.com', password='somesupersecret')
                >>> client.accounts.get_profile()
                {
                    "success": True,
                    "resource": {
                        "id": "507f1f77bcf86cd799439011",
                        "email": "test@test.com"
                        "name": "Test"
                        "surname": "McTestyFace"
                        "company": "Acme"
                        "avatar": "https://encrypted-tbn0.gstatic.com/images?q=tbn:ANd9GcRDzvciiQ5-P6itEjycriGb9l9sJH9E538C-tM9QRgFVTaj0Muq"
                        "role": "user"
                    }
                }
        """
        return self.make_request('get_profile', '/')

    def update_profile(self, data):
        """Update the current logged in user's profile
        
        Arguments:
            data {dict} -- A dictionary of profile values to be updated
        
        Returns:
            dict -- A confirmfation payload of the updated keys

        Example:
            .. code-block:: python

                >>> from speckle.SpeckleClient import SpeckleApiClient
                >>> client = SpeckleApiClient(host='hestia.speckle.com')
                >>> client.login(email='test@test.com', password='somesupersecret')
                >>> client.accounts.get_profile()
                {
                    "success": True,
                    "resource": {
                        "id": "507f1f77bcf86cd799439011",
                        "email": "test@test.com"
                        "name": "Test"
                        "surname": "McTestyFace"
                        "company": "Acme"
                        "avatar": "https://encrypted-tbn0.gstatic.com/images?q=tbn:ANd9GcRDzvciiQ5-P6itEjycriGb9l9sJH9E538C-tM9QRgFVTaj0Muq"
                        "role": "user"
                    }
                }
                >>> client.accounts.update_profile({'name': 'Tester'})
                { 
                    "success": True,
                    "message": 'User profile updated.'
                }

        """
        return self.make_request('update_profile', '/', data)

    def set_role(self, id, role):
        """Set the role of a user

        Warning:
            The client must be authenticated with an `admin` user to carry
            out this operation.
        
        Arguments:
            id {str} -- The ID of the user to be updated
            role {str} -- The role to give the user ('admin' or 'user')
        
        Returns:
            dict -- A response payload confirming the user role update

        Example:
            .. code-block:: python

                >>> from speckle.SpeckleClient import SpeckleApiClient
                >>> client = SpeckleApiClient(host='hestia.speckle.com')
                >>> client.login(email='test@test.com', password='somesupersecret')
                >>> client.accounts.get_profile()
                {
                    "success": True,
                    "resource": {
                        "id": "507f1f77bcf86cd799439011",
                        "email": "test@test.com"
                        "name": "Test"
                        "surname": "McTestyFace"
                        "company": "Acme"
                        "avatar": "https://encrypted-tbn0.gstatic.com/images?q=tbn:ANd9GcRDzvciiQ5-P6itEjycriGb9l9sJH9E538C-tM9QRgFVTaj0Muq"
                        "role": "admin"
                    }
                }
                >>> user_to_modify = '9eeb71aa8b2a292512a3bf94091e2df8'
                >>> client.accounts.set_role(id=user_to_modify, role='admin')
                { 
                    "success": True,
                    "message": 'User profile updated.'
                }
        """
        data = {'role': role}
        return self.make_request('set_role', '/' + id, data)

    def search(self, search):
        """Search for one or more users
        
        Arguments:
            search {str} -- A search string. Should be at least 3 characters long.

        Returns:
            list -- A list of found users

        Example:
            .. code-block:: python

                >>> client.accounts.search('tom')
                {
                    "success": True,
                    "resources": [{
                        "id": "507f1f77bcf86cd799439011",
                        "email": "tom@test.com"
                        "name": "Tom"
                        "surname": "Svilans"
                        "company": "Acme"
                        "avatar": "https://encrypted-tbn0.gstatic.com/images?q=tbn:ANd9GcRDzvciiQ5-P6itEjycriGb9l9sJH9E538C-tM9QRgFVTaj0Muq"
                        "role": "admin"
                    },
                    {
                        "id": "397f19073b30eb9f71fc5bas",
                        "email": "test@tom.com"
                        "name": "Test"
                        "surname": "McTestyFace"
                        "company": "Tom Associated"
                        "avatar": "https://encrypted-tbn0.gstatic.com/images?q=tbn:ANd9GcRDzvciiQ5-P6itEjycriGb9l9sJH9E538C-tM9QRgFVTaj0Muq"
                        "role": "user"
                    },
                    {
                        "id": "9bfb9894c21e29014cd0e8166",
                        "email": "tom@tom.com"
                        "name": "Tom"
                        "surname": "Tom"
                        "company": "Tom"
                        "avatar": "https://i.ytimg.com/vi/Dtm5nMlqq1Q/maxresdefault.jpg"
                        "role": "admin"
                    }]

                }
        """
        data = {
            'searchString': search
        }

        return self.make_request('search', 'search', data)



import uuid
from pydantic import BaseModel, UUID4, validator, Schema
from typing import List, Optional
from speckle.base.resource import ResourceBase, ResourceBaseSchema
from speckle.resources.objects import SpeckleObject
from speckle.resources.api_clients import ApiClient

NAME = 'streams'
METHODS = ['list', 'create', 'get', 'update',
           'delete', 'comment_get', 'comment_create',
           'clone', 'diff', 'list_objects', 'list_clients']


class LayerProperties(BaseModel):
  color: dict = {}
  visible: bool = True
  pointsize: float = 0
  linewidth: float = 0
  shininess: float = 0
  smooth: bool = True
  showEdges: bool = True
  wireframe: bool = True


class Layer(BaseModel):
    guid: str
    name: str = 'New Layer'
    orderIndex: int = 0
    startIndex: int = 0
    objectCount: int = 0
    topology: Optional[str]
    properties: Optional[LayerProperties]

    @validator('guid', pre=True, always=True)
    def set_guid(cls, v):
        return v or str(uuid.uuid4())


class Stream(ResourceBaseSchema):
    streamId: Optional[str]
    name: Optional[str]
    description: Optional[str]
    tags: List[str] = []
    commitMessage: str  = 'Modified stream'
    objects: List[SpeckleObject] = []
    layers: List[Layer] = []
    parent: Optional[str]
    children: List[str] = []

class Resource(ResourceBase):
    """API Access class for Streams

    """
    
    def __init__(self, session, basepath, me):
        super().__init__(session, basepath, me, NAME, METHODS)

        self.method_dict.update({
            'clone': {
                'method': 'POST'
            },
            'diff': {
                'method': 'GET'
            },
            'list_objects': {
                'method': 'GET'
            },
            'list_clients': {
                'method': 'GET'
            }
        })

        self.schema = Stream

    def list(self):
        """List all streams
        
        Returns:
            list -- A list of Streams, without objects attached
        """
        return self.make_request('list', '?omit=objects')

    def create(self, data):
        """Create a stream from a data dictionary
        
        Arguments:
            data {dict} -- A dictionary describing a stream
        
        Returns:
            Stream -- The instance created on the Speckle Server
        """
        return self.make_request('create', '/', data)

    def get(self, id):
        """Get a specific stream from the SpeckleServer
        
        Arguments:
            id {str} -- The StreamId of the stream to retrieve
        
        Returns:
            Stream -- The stream
        """
        return self.make_request('get', '/' + id)

    def update(self, id, data):
        """Update a specific stream
        
        Arguments:
            id {str} -- The StreamId of the stream to update
            data {dict} -- A dict of values to update
        
        Returns:
            dict -- a confirmation payload with the updated keys
        """
        return self.make_request('update', '/' + id, data)

    def delete(self, id):
        """Delete a specific stream
        
        Arguments:
            id {str} -- The StreamId of the stream to delete
        
        Returns:
            dict -- A confirmation payload
        """
        return self.make_request('delete', '/' + id)

    def comment_get(self, id):
        """Retrieve comments attached to a stream
        
        Arguments:
            id {str} -- The StreamId of the stream to retrieve comments from
        
        Returns:
            list -- A list of comments
        """
        return self.make_request('comment_get', '/' + id, comment=True)

    def comment_create(self, id, data):
        """Add a comment to a stream
        
        Arguments:
            id {str} -- The StreamId of the stream to comment on
            data {dict} -- A comment dictionary object
        
        Returns:
            CommentSchema -- The comment created by the server
        """
        return self.make_request('comment_create', '/' + id, data, comment=True)


    def clone(self, id, name=None):
        """Clone a stream
        
        Arguments:
            id {str} -- The StreamId of the stream to clone
        
        Keyword Arguments:
            name {str} -- The name of the new cloned stream. eg: stream-x-2019-06-09-backup (default: {None})
        
        Returns:
            tuple -- The clone and parent stream as dicts
        """
        response = self.make_request('clone', '/' + id + '/clone', {'name': name})
        clone = self._parse_response(response['clone'])
        parent = self._parse_response(response['parent'])
        return clone, parent

    def diff(self, id, other_id):
        """Runs a diff on two streams
        
        Arguments:
            id {str} -- StreamId of the main stream
            other_id {str} -- StreamId of the stream to compare
        
        Returns:
            dict -- A response payload with objects and layers as keys
        """
        return self.make_request('diff', '/' + id + '/diff/' + other_id)

    def list_objects(self, id):
        """Return the list of objects in a stream
        
        Arguments:
            id {str} -- StreamId of the stream to list objects from
        
        Returns:
            list -- A list of Speckle objects
        """
        return self.make_request('list_objects', '/' + id + '/objects', schema=SpeckleObject)

    def list_clients(self, id):
        """Return the list of api clients connected to the stream
        
        Arguments:
            id {str} -- StreamId of the stream to list objects from
        
        Returns:
            list -- A list of API clients
        """
        return self.make_request('list_clients', '/' + id + '/clients', schema=ApiClient)
